import React, { useEffect, useState, useRef, useId } from "react";

export interface CodeInputProps {
  length?: number;
  charset?: Charset;
  value?: string;
  onChange?: Function;
  pasteParser?: Function;
  className?: string;
  disabled?: boolean;
  autoFocus?: boolean;
  children: any;
}

const Key = {
  BACKSPACE: "Backspace",
  ARROW_LEFT: "ArrowLeft",
  ARROW_RIGHT: "ArrowRight",
} as const;

const charsetRegex = {
  alpha: /[a-zA-Z]/,
  alphanumeric: /[a-zA-Z0-9]/,
  numeric: /[0-9]/,
} as const;

type Charset = keyof typeof charsetRegex;

export const CodeInput = ({
  length = 4,
  charset = "numeric",
  onChange,
  value,
  pasteParser,
  disabled,
  autoFocus,
  children,
}: CodeInputProps) => {
  const id = useId();

  const [chars, setChars] = useState(() => value?.split("") || []);
  const inputsRef = useRef<(HTMLInputElement | null)[]>([]);

  // Callback for on change
  useEffect(() => {
    const code = chars.join("");
    onChange?.call(null, { target: { value: code } });
  }, [chars]);

  // Handles paste event
  function handlePaste(e: React.ClipboardEvent) {
    const pastedText = e.clipboardData.getData("text");

    const parsed = (pasteParser?.call(null, pastedText) || pastedText)
      .slice(0, length)
      .split("");

    setChars(parsed);

    inputsRef.current[length - 1]?.focus();
  }

  function move(direction: "left" | "right", index: number) {
    const newIndex = direction === "left" ? index - 1 : index + 1;
    const inputEl = inputsRef.current[newIndex];

    if (inputEl) {
      inputEl.focus();
      inputEl.setSelectionRange(0, 1);
    }
  }

  function handleKey(key: string, inputIndex: number) {
    const char = key;

    switch (key) {
      case Key.ARROW_LEFT:
        move("left", inputIndex);
        break;

      case Key.ARROW_RIGHT:
        move("right", inputIndex);
        break;

      case Key.BACKSPACE:
        if (inputsRef.current[inputIndex]?.value === "")
          move("left", inputIndex);
        else
          setChars((oldChars) => {
            const chars = [...oldChars];
            chars[inputIndex] = "";
            return chars;
          });
        break;

      default:
        if (char.match(charsetRegex[charset])) {
          setChars((oldChars) => {
            const chars = [...oldChars];
            chars[inputIndex] = char;
            return chars;
          });

          move("right", inputIndex);
        }
    }
  }

  function handleKeyDown(e: React.KeyboardEvent) {
    if (e.metaKey || e.key === "Unidentified" || e.key === "Tab") return;
    e.preventDefault();

    const inputIndex = +(e.currentTarget.id.split("_").at(-1) as string);
    handleKey(e.key, inputIndex);
  }

  // To make it work on mobiles
  function handleInput(e: React.FormEvent<HTMLInputElement>) {
    e.preventDefault();

    let key = e.currentTarget.value;
    key = key === "" ? "Backspace" : key;

    const inputIndex = +(e.currentTarget.id.split("_").at(-1) as string);
    handleKey(key, inputIndex);
  }

  return (
    <div className='flex gap-2'>
      {[...Array(length)].map((_, i) =>
        React.cloneElement(children, {
          autoFocus: autoFocus && i === 0,
          key: `${id}_${i}`,
          id: `${id}_${i}`,
          maxLength: 1,
          ref: (el: HTMLInputElement): void => {
            inputsRef.current[i] = el;
          },
          disabled,
          value: chars[i] ?? "",
          onPaste: handlePaste,
          onInput: handleInput,
          onKeyDown: handleKeyDown,
          onChange: () => {},
        })
      )}
    </div>
  );
};
